from .variational_autoencoder import VariationalAutoEncoder
from keras.layers import Input, Dense, ReLU
from keras.models import Model


class SmallDenseVAE(VariationalAutoEncoder):

    def _encoder(self):
        input_tensor = Input(shape=self.input_shape[1:])
        x = Dense(420)(input_tensor)
        x = ReLU()(x)
        x = Dense(210)(x)
        x = ReLU()(x)
        x = Dense(105)(x)
        x = ReLU()(x)
        encoder = Model(input_tensor, x)
        return encoder

    def _decoder(self):
        latent = Input(shape=(self.latent_dim,))
        x = Dense(200)(latent)
        x = ReLU()(x)
        x = Dense(400)(x)
        x = ReLU()(x)
        x = Dense(800)(x)
        x = ReLU()(x)
        reco = Dense(self.input_shape[1])(x)
        decoder = Model(latent, reco)
        return decoder



from keras.preprocessing.image import ImageDataGenerator
import numpy as np
from .data_loader import DataLoader


class DataGenerator:
    def __init__(self, image_shape, batch_size, file_path='/data/scramtex_700_28px.pkl', binarize=False):
        self.DATA_LOADER = DataLoader(file_path, binarize)
        self.IMAGE_SHAPE = image_shape
        self.BATCH_SIZE = batch_size

        self.DATA_GENERATOR = ImageDataGenerator()

        self.TRAIN, self.TEST = self.DATA_LOADER.train_test_split()

    def flow(self):
        return self.DATA_GENERATOR.flow(self.TRAIN, self.TRAIN, batch_size=self.BATCH_SIZE)

    def contrast_flow(self):
        def train_generator(_it):
            while True:
                batch_x, batch_y = next(_it)
                contrast = np.random.rand()
                _batch_x, _batch_y = contrast*batch_x, contrast*batch_y
                _batch_x, _batch_y = np.clip(_batch_x, 0, 1), np.clip(_batch_y, 0, 1)
                yield _batch_x, _batch_y
        return train_generator(self.flow())

    def flattened_flow(self):
        def train_generator(_it):
            image_dim = np.prod(self.IMAGE_SHAPE)
            while True:
                batch_x, batch_y = next(_it)
                yield batch_x.reshape(self.BATCH_SIZE, image_dim), batch_y.reshape(self.BATCH_SIZE, image_dim)
        return train_generator(self.flow())

    def _contrast(self, img, alpha, brightness=0):
        return np.clip(alpha*(img - 0.5) + 0.5 + brightness, 0, 1)

    def flattened_contrast_flow(self):
        def train_generator(_it):
            image_dim = np.prod(self.IMAGE_SHAPE)
            contrast = np.vectorize(self._contrast)
            while True:
                batch_x, batch_y = next(_it)
                """
                    Tested contrast function with random alpha value between 0-2!
                """
                alpha = np.random.rand(batch_x.shape[0])*2.
                _batch_x, _batch_y = contrast(batch_x, alpha), contrast(batch_y, alpha)
                yield batch_x.reshape(self.BATCH_SIZE, image_dim), batch_y.reshape(self.BATCH_SIZE, image_dim)
        return train_generator(self.flow())

    def validation_data(self):
        return self.TEST, self.TEST

    def flattened_validation_data(self):
        image_dim = np.prod(self.IMAGE_SHAPE)
        reshaped_test = self.TEST.reshape(self.TEST.shape[0], image_dim)
        return reshaped_test, reshaped_test



import matplotlib.pyplot as plt
from keras.models import load_model
import os


class ModelTrainer:
    def __init__(self, model, data_generator, loss_fn="normal", lr=1e-3, decay=5e-5, observation_noise=1e-3, beta=1e-12):
        assert loss_fn in str(["normal", "normalDiag", "binary", None]),\
            "Loss function should be in [\'normal\', \'normalDiag\', \'None\' (bernoulli), \'binary\']"
        args = loss_fn, lr, decay, observation_noise, beta
        self._model = model
        self.model, self.generator = self._model.get_compiled_model(*args)
        self.latent_model = self._model.latent_model
        self.latent_dim = model.latent_dim
        self.data_generator = data_generator
        self.saved = False
        self.model.summary()

    def fit(self, EPOCHS, STEPS, contrast=False):
        try:
            self.history = self.model.fit_generator(
                self.data_generator.flow() if not contrast else self.data_generator.contrast_flow(),
                steps_per_epoch=STEPS,
                verbose=1, epochs=EPOCHS,
                validation_data=self.data_generator.validation_data())
        except ValueError:
            self.history = self.model.fit_generator(
                self.data_generator.flattened_flow() if not contrast else self.data_generator.flattened_contrast_flow(), steps_per_epoch=STEPS,
                verbose=1, epochs=EPOCHS,
                validation_data=self.data_generator.flattened_validation_data())
        finally:
            self._save_model()
            plt.title("Model loss")
            plt.plot(self.history.history['loss'])
            plt.plot(self.history.history['val_loss'])
            plt.ylabel('Loss')
            plt.xlabel('Epoch')
            plt.legend(['train', 'validation'], loc='upper right')
            plt.show()

    def _save_model(self):
        print("Saving the trained inference, generator and latent models...\t", end='')
        self.model.save(os.getcwd() + "/model.h5")
        self.generator.save(os.getcwd() + "/generator_model.h5")
        """
            This must be acquired here before we are after training at this point!
        """
        self.latent_model.save(os.getcwd() + "/latent_model.h5")
        self.saved = True
        print("OK!")

    def load_models(self):
        if not self.saved:
            print("The models have not been trained or saved yet.")
            return
        print("Loading models...\t", end='')
        self.model = load_model(os.getcwd() + "/model.h5")
        self.generator = load_model(os.getcwd() + "/generator-model.h5")
        self.latent_model = load_model(os.getcwd() + "/latent_model.h5")
        print("OK!")
